'''
CA: 데이터 버퍼링을 추가한 향상된 버전을 설명드리겠습니다. 버퍼링은 데이터의 일시적인 손실을 방지하고 더 부드러운 디스플레이를 가능하게 합니다.
'''

import sys
import socket
import json
import numpy as np
from collections import deque

# PyQt5
# from PyQt5.QtWidgets import (QApplication, QMainWindow, QVBoxLayout, QWidget, 
#                            QHBoxLayout, QLabel, QSpinBox, QPushButton)
# from PyQt5.QtCore import QTimer

# PyQt6
from PyQt6.QtWidgets import (QApplication, QMainWindow, QVBoxLayout, QWidget, 
                           QHBoxLayout, QLabel, QSpinBox, QPushButton)
from PyQt6.QtCore import QTimer

from pyqtgraph import PlotWidget

class BufferedFFTVisualizer(QMainWindow):
    def __init__(self, port=5000, buffer_size=10):
        super().__init__()
        self.setWindowTitle('Buffered FFT Spectrum Analyzer')
        self.setGeometry(100, 100, 1000, 700)

        # 버퍼 설정
        self.buffer_size = buffer_size
        self.data_buffer = deque(maxlen=buffer_size)
        self.frequencies = None
        self.current_average = None
        
        self.setup_ui()
        self.setup_network(port)
        self.setup_timers()

    def setup_ui(self):
        # 메인 위젯과 레이아웃
        main_widget = QWidget()
        self.setCentralWidget(main_widget)
        layout = QVBoxLayout(main_widget)
        
        # 컨트롤 패널
        control_panel = QWidget()
        control_layout = QHBoxLayout(control_panel)
        layout.addWidget(control_panel)
        
        # 버퍼 크기 컨트롤
        buffer_label = QLabel("Buffer Size:")
        self.buffer_spinbox = QSpinBox()
        self.buffer_spinbox.setRange(1, 100)
        self.buffer_spinbox.setValue(self.buffer_size)
        self.buffer_spinbox.valueChanged.connect(self.change_buffer_size)
        
        # 평균화 모드 토글 버튼
        self.average_button = QPushButton("Toggle Averaging")
        self.average_button.setCheckable(True)
        self.average_button.clicked.connect(self.toggle_averaging)
        
        # 버퍼 클리어 버튼
        clear_button = QPushButton("Clear Buffer")
        clear_button.clicked.connect(self.clear_buffer)
        
        # 컨트롤 패널에 위젯 추가
        control_layout.addWidget(buffer_label)
        control_layout.addWidget(self.buffer_spinbox)
        control_layout.addWidget(self.average_button)
        control_layout.addWidget(clear_button)
        control_layout.addStretch()
        
        # 플롯 위젯 설정
        self.plot_widget = PlotWidget()
        layout.addWidget(self.plot_widget)
        
        # 플롯 설정
        self.setup_plot()
        
        # 상태 표시줄 추가
        self.statusBar().showMessage('Waiting for connection...')

    def setup_plot(self):
        self.plot_widget.setBackground('white')
        self.plot_widget.showGrid(x=True, y=True)
        self.plot_widget.setLabel('left', 'Amplitude')
        self.plot_widget.setLabel('bottom', 'Frequency (Hz)')
        self.plot_widget.setTitle('Real-time FFT Spectrum with Buffering', 
                                color='black', size='12pt')
        
        # 현재 데이터와 평균 데이터를 위한 두 개의 곡선
        self.current_curve = self.plot_widget.plot(pen='b', name='Current')
        self.average_curve = self.plot_widget.plot(pen='r', name='Average')
        
        # 범례 추가
        self.plot_widget.addLegend()

    def setup_network(self, port):
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind(('localhost', port))
        self.server_socket.listen(1)
        self.server_socket.setblocking(False)
        
        self.client_socket = None
        self.connected = False

    def setup_timers(self):
        # 연결 타이머
        self.connect_timer = QTimer()
        self.connect_timer.timeout.connect(self.try_connect)
        self.connect_timer.start(100)
        
        # 업데이트 타이머
        self.update_timer = QTimer()
        self.update_timer.timeout.connect(self.update_plot)
        self.update_timer.start(50)

    def try_connect(self):
        if not self.connected:
            try:
                self.client_socket, _ = self.server_socket.accept()
                self.client_socket.setblocking(False)
                self.connected = True
                self.statusBar().showMessage('Connected!')
                self.connect_timer.stop()
            except BlockingIOError:
                pass
            except Exception as e:
                self.statusBar().showMessage(f'Connection error: {e}')

    def update_plot(self):
        if not self.connected:
            return

        try:
            # data = self.client_socket.recv(4096).decode()

            k_buff_size = 4096
            data = b''
            while True:
                part = self.client_socket.recv(k_buff_size)
                data += part
                if len(part) < k_buff_size:
                    break
            data = data.decode()

            if data:
                fft_data = json.loads(data)
                frequencies = np.array(fft_data['frequencies'])
                amplitudes = np.array(fft_data['amplitudes'])
                
                # 첫 데이터 수신시 주파수 배열 저장
                if self.frequencies is None:
                    self.frequencies = frequencies
                
                # 버퍼에 데이터 추가
                self.data_buffer.append(amplitudes)
                
                # 현재 데이터 표시
                self.current_curve.setData(frequencies, amplitudes)
                
                # 평균 모드가 활성화된 경우 평균 계산 및 표시
                if self.average_button.isChecked():
                    avg_data = np.mean(self.data_buffer, axis=0)
                    self.average_curve.setData(frequencies, avg_data)
                else:
                    self.average_curve.clear()
                
                # 상태 표시줄 업데이트
                self.statusBar().showMessage(
                    f'Buffer: {len(self.data_buffer)}/{self.buffer_size} | '
                    f'Averaging: {self.average_button.isChecked()}'
                )
                
        except BlockingIOError:
            pass
        except Exception as e:
            self.statusBar().showMessage(f'Error during update: {e}')
            self.connected = False
            self.connect_timer.start()

    def change_buffer_size(self, new_size):
        # 새로운 버퍼 생성 및 이전 데이터 복사
        new_buffer = deque(self.data_buffer, maxlen=new_size)
        self.data_buffer = new_buffer
        self.buffer_size = new_size

    def toggle_averaging(self, checked):
        if not checked:
            self.average_curve.clear()

    def clear_buffer(self):
        self.data_buffer.clear()
        self.average_curve.clear()
        self.statusBar().showMessage('Buffer cleared')

    def closeEvent(self, event):
        if self.client_socket:
            self.client_socket.close()
        self.server_socket.close()
        event.accept()

def main():
    app = QApplication(sys.argv)
    app.setStyle('Fusion')
    window = BufferedFFTVisualizer()
    window.show()

    # PyQt5
    # sys.exit(app.exec_())

    # PyQt6
    sys.exit(app.exec())

if __name__ == '__main__':
    main()

'''
CA: 이 향상된 버전의 주요 특징을 설명드리겠습니다:

데이터 버퍼링 기능:

설정 가능한 버퍼 크기
이동 평균 계산 기능
버퍼 내용 실시간 시각화


새로운 UI 컨트롤:

버퍼 크기 조절 스핀박스
평균화 모드 토글 버튼
버퍼 클리어 버튼
상태 표시줄에 현재 버퍼 상태 표시


향상된 시각화:

현재 데이터와 평균 데이터를 다른 색상으로 표시
범례 추가
그리드 표시


버퍼링의 장점:

노이즈 감소
일시적인 데이터 손실 방지
트렌드 분석 용이
평균화를 통한 안정적인 표시



이 구현을 사용하기 위해서는 기존과 동일한 패키지들이 필요합니다.
다음으로 그래프 설정(축 범위, 스케일 등) 커스터마이징에 대해 설명드릴까요?

HS: 코드를 나중에 실행해서 확인해 보겠습니다. 그리고 뭔가가 잘 안 되면 그때 다시 문의하겠습니다. 
우선은 다음 설명을 부탁합니다.

'''